import pytest

from distributed.deploy.cluster import Cluster
from distributed.utils_test import cleanup  # noqa: F401


@pytest.mark.asyncio
async def test_eq(cleanup):
    clusterA = Cluster(asynchronous=True, name="A")
    clusterA2 = Cluster(asynchronous=True, name="A2")
    clusterB = Cluster(asynchronous=True, name="B")
    assert clusterA != "A"
    assert not (clusterA == "A")
    assert clusterA == clusterA
    assert not (clusterA != clusterA)
    assert clusterA != clusterA2
    assert not (clusterA == clusterA2)
    assert clusterA != clusterB
    assert not (clusterA == clusterB)


@pytest.mark.asyncio
async def test_repr(cleanup):
    cluster = Cluster(asynchronous=True, name="A")
    assert cluster.scheduler_address == "<Not Connected>"
    res = repr(cluster)
    expected = "Cluster(A, '<Not Connected>', workers=0, threads=0, memory=0 B)"
    assert res == expected



import asyncio
import functools
import threading
from queue import Queue

from .client import Future, default_client
from .protocol import to_serialize
from .utils import iscoroutinefunction, sync, thread_state
from .utils_comm import WrappedKey
from .worker import get_worker


class Actor(WrappedKey):
    """Controls an object on a remote worker

    An actor allows remote control of a stateful object living on a remote
    worker.  Method calls on this object trigger operations on the remote
    object and return ActorFutures on which we can block to get results.

    Examples
    --------
    >>> class Counter:
    ...    def __init__(self):
    ...        self.n = 0
    ...    def increment(self):
    ...        self.n += 1
    ...        return self.n

    >>> from dask.distributed import Client
    >>> client = Client()

    You can create an actor by submitting a class with the keyword
    ``actor=True``.

    >>> future = client.submit(Counter, actor=True)
    >>> counter = future.result()
    >>> counter
    <Actor: Counter, key=Counter-1234abcd>

    Calling methods on this object immediately returns deferred ``ActorFuture``
    objects.  You can call ``.result()`` on these objects to block and get the
    result of the function call.

    >>> future = counter.increment()
    >>> future.result()
    1
    >>> future = counter.increment()
    >>> future.result()
    2
    """

    def __init__(self, cls, address, key, worker=None):
        self._cls = cls
        self._address = address
        self.key = key
        self._future = None
        if worker:
            self._worker = worker
            self._client = None
        else:
            try:
                self._worker = get_worker()
            except ValueError:
                self._worker = None
            try:
                self._client = default_client()
                self._future = Future(key)
            except ValueError:
                self._client = None

    def __repr__(self):
        return "<Actor: %s, key=%s>" % (self._cls.__name__, self.key)

    def __reduce__(self):
        return (Actor, (self._cls, self._address, self.key))

    @property
    def _io_loop(self):
        if self._worker:
            return self._worker.io_loop
        else:
            return self._client.io_loop

    @property
    def _scheduler_rpc(self):
        if self._worker:
            return self._worker.scheduler
        else:
            return self._client.scheduler

    @property
    def _worker_rpc(self):
        if self._worker:
            return self._worker.rpc(self._address)
        else:
            if self._client.direct_to_workers:
                return self._client.rpc(self._address)
            else:
                return ProxyRPC(self._client.scheduler, self._address)

    @property
    def _asynchronous(self):
        if self._client:
            return self._client.asynchronous
        else:
            return threading.get_ident() == self._worker.thread_id

    def _sync(self, func, *args, **kwargs):
        if self._client:
            return self._client.sync(func, *args, **kwargs)
        else:
            # TODO support sync operation by checking against thread ident of loop
            return sync(self._worker.loop, func, *args, **kwargs)

    def __dir__(self):
        o = set(dir(type(self)))
        o.update(attr for attr in dir(self._cls) if not attr.startswith("_"))
        return sorted(o)

    def __getattr__(self, key):

        if self._future and self._future.status not in ("finished", "pending"):
            raise ValueError(
                "Worker holding Actor was lost.  Status: " + self._future.status
            )

        if (
            self._worker
            and self._worker.address == self._address
            and getattr(thread_state, "actor", False)
        ):
            # actor calls actor on same worker
            actor = self._worker.actors[self.key]
            attr = getattr(actor, key)

            if iscoroutinefunction(attr):
                return attr

            elif callable(attr):
                return lambda *args, **kwargs: ActorFuture(
                    None, None, result=attr(*args, **kwargs)
                )
            else:
                return attr

        attr = getattr(self._cls, key)

        if callable(attr):

            @functools.wraps(attr)
            def func(*args, **kwargs):
                async def run_actor_function_on_worker():
                    try:
                        result = await self._worker_rpc.actor_execute(
                            function=key,
                            actor=self.key,
                            args=[to_serialize(arg) for arg in args],
                            kwargs={k: to_serialize(v) for k, v in kwargs.items()},
                        )
                    except OSError:
                        if self._future:
                            await self._future
                        else:
                            raise OSError("Unable to contact Actor's worker")
                    return result["result"]

                if self._asynchronous:
                    return asyncio.ensure_future(run_actor_function_on_worker())
                else:
                    # TODO: this mechanism is error prone
                    # we should endeavor to make dask's standard code work here
                    q = Queue()

                    async def wait_then_add_to_queue():
                        x = await run_actor_function_on_worker()
                        q.put(x)

                    self._io_loop.add_callback(wait_then_add_to_queue)

                    return ActorFuture(q, self._io_loop)

            return func

        else:

            async def get_actor_attribute_from_worker():
                x = await self._worker_rpc.actor_attribute(
                    attribute=key, actor=self.key
                )
                return x["result"]

            return self._sync(get_actor_attribute_from_worker)

    @property
    def client(self):
        return self._future.client


class ProxyRPC:
    """
    An rpc-like object that uses the scheduler's rpc to connect to a worker
    """

    def __init__(self, rpc, address):
        self.rpc = rpc
        self._address = address

    def __getattr__(self, key):
        async def func(**msg):
            msg["op"] = key
            result = await self.rpc.proxy(worker=self._address, msg=msg)
            return result

        return func


class ActorFuture:
    """Future to an actor's method call

    Whenever you call a method on an Actor you get an ActorFuture immediately
    while the computation happens in the background.  You can call ``.result``
    to block and collect the full result

    See Also
    --------
    Actor
    """

    def __init__(self, q, io_loop, result=None):
        self.q = q
        self.io_loop = io_loop
        if result:
            self._cached_result = result

    def __await__(self):
        return self.result()

    def result(self, timeout=None):
        try:
            return self._cached_result
        except AttributeError:
            self._cached_result = self.q.get(timeout=timeout)
            return self._cached_result

    def __repr__(self):
        return "<ActorFuture>"



import asyncio
import logging
import uuid
import warnings
from asyncio import TimeoutError
from collections import defaultdict, deque

from tornado.ioloop import IOLoop, PeriodicCallback

import dask

from distributed.utils_comm import retry_operation

from .metrics import time
from .utils import log_errors, parse_timedelta, sync, thread_state
from .worker import get_client, get_worker

logger = logging.getLogger(__name__)


class _Watch:
    def __init__(self, duration=None):
        self.duration = duration
        self.started_at = None

    def start(self):
        self.started_at = time()

    def elapsed(self):
        return time() - self.started_at

    def leftover(self):
        if self.duration is None:
            return None
        else:
            return max(0, self.duration - self.elapsed())


class SemaphoreExtension:
    """An extension for the scheduler to manage Semaphores

    This adds the following routes to the scheduler

    * semaphore_acquire
    * semaphore_release
    * semaphore_close
    * semaphore_refresh_leases
    * semaphore_register
    """

    def __init__(self, scheduler):
        self.scheduler = scheduler

        # {semaphore_name: asyncio.Event}
        self.events = defaultdict(asyncio.Event)
        # {semaphore_name: max_leases}
        self.max_leases = dict()
        # {semaphore_name: {lease_id: lease_last_seen_timestamp}}
        self.leases = defaultdict(dict)

        self.scheduler.handlers.update(
            {
                "semaphore_register": self.create,
                "semaphore_acquire": self.acquire,
                "semaphore_release": self.release,
                "semaphore_close": self.close,
                "semaphore_refresh_leases": self.refresh_leases,
                "semaphore_value": self.get_value,
            }
        )

        self.scheduler.extensions["semaphores"] = self

        # {metric_name: {semaphore_name: metric}}
        self.metrics = {
            "acquire_total": defaultdict(int),  # counter
            "release_total": defaultdict(int),  # counter
            "average_pending_lease_time": defaultdict(float),  # gauge
            "pending": defaultdict(int),  # gauge
        }

        validation_callback_time = parse_timedelta(
            dask.config.get("distributed.scheduler.locks.lease-validation-interval"),
            default="s",
        )
        self._pc_lease_timeout = PeriodicCallback(
            self._check_lease_timeout, validation_callback_time * 1000
        )
        self._pc_lease_timeout.start()
        self.lease_timeout = parse_timedelta(
            dask.config.get("distributed.scheduler.locks.lease-timeout"), default="s"
        )

    async def get_value(self, comm=None, name=None):
        return len(self.leases[name])

    # `comm` here is required by the handler interface
    def create(self, comm=None, name=None, max_leases=None):
        # We use `self.max_leases` as the point of truth to find out if a semaphore with a specific
        # `name` has been created.
        if name not in self.max_leases:
            assert isinstance(max_leases, int), max_leases
            self.max_leases[name] = max_leases
        else:
            if max_leases != self.max_leases[name]:
                raise ValueError(
                    "Inconsistent max leases: %s, expected: %s"
                    % (max_leases, self.max_leases[name])
                )

    def refresh_leases(self, comm=None, name=None, lease_ids=None):
        with log_errors():
            now = time()
            logger.debug(
                "Refresh leases for %s with ids %s at %s", name, lease_ids, now
            )
            for id_ in lease_ids:
                if id_ not in self.leases[name]:
                    logger.critical(
                        f"Refreshing an unknown lease ID {id_} for {name}. This might be due to leases "
                        f"timing out and may cause overbooking of the semaphore!"
                        f"This is often caused by long-running GIL-holding in the task which acquired the lease."
                    )
                self.leases[name][id_] = now

    def _get_lease(self, name, lease_id):
        result = True

        if (
            # This allows request idempotency
            lease_id in self.leases[name]
            or len(self.leases[name]) < self.max_leases[name]
        ):
            now = time()
            logger.debug("Acquire lease %s for %s at %s", lease_id, name, now)
            self.leases[name][lease_id] = now
            self.metrics["acquire_total"][name] += 1
        else:
            result = False
        return result

    def _semaphore_exists(self, name):
        if name not in self.max_leases:
            return False
        return True

    async def acquire(self, comm=None, name=None, timeout=None, lease_id=None):
        with log_errors():
            if not self._semaphore_exists(name):
                raise RuntimeError(f"Semaphore `{name}` not known or already closed.")

            if isinstance(name, list):
                name = tuple(name)
            w = _Watch(timeout)
            w.start()

            self.metrics["pending"][name] += 1
            while True:
                logger.debug(
                    "Trying to acquire %s for %s with %s seconds left.",
                    lease_id,
                    name,
                    w.leftover(),
                )
                # Reset the event and try to get a release. The event will be set if the state
                # is changed and helps to identify when it is worth to retry an acquire
                self.events[name].clear()

                result = self._get_lease(name, lease_id)

                # If acquiring fails, we wait for the event to be set, i.e. something has
                # been released and we can try to acquire again (continue loop)
                if not result:
                    future = asyncio.wait_for(
                        self.events[name].wait(), timeout=w.leftover()
                    )
                    try:
                        await future
                        continue
                    except TimeoutError:
                        result = False
                logger.debug(
                    "Acquisition of lease %s for %s is %s after waiting for %ss.",
                    lease_id,
                    name,
                    result,
                    w.elapsed(),
                )
                # We're about to return, so the lease is no longer "pending"
                self.metrics["average_pending_lease_time"][name] = (
                    self.metrics["average_pending_lease_time"][name] + w.elapsed()
                ) / 2
                self.metrics["pending"][name] -= 1

                return result

    def release(self, comm=None, name=None, lease_id=None):
        with log_errors():
            if not self._semaphore_exists(name):
                logger.warning(
                    f"Tried to release semaphore `{name}` but it is not known or already closed."
                )
                return
            if isinstance(name, list):
                name = tuple(name)
            if name in self.leases and lease_id in self.leases[name]:
                self._release_value(name, lease_id)
            else:
                logger.warning(
                    f"Tried to release semaphore but it was already released: "
                    f"name={name}, lease_id={lease_id}. This can happen if the semaphore timed out before."
                )

    def _release_value(self, name, lease_id):
        logger.debug("Releasing %s for %s", lease_id, name)
        # Everything needs to be atomic here.
        del self.leases[name][lease_id]
        self.events[name].set()
        self.metrics["release_total"][name] += 1

    def _check_lease_timeout(self):
        now = time()
        semaphore_names = list(self.leases.keys())
        for name in semaphore_names:
            ids = list(self.leases[name])
            logger.debug(
                "Validating leases for %s at time %s. Currently known %s",
                name,
                now,
                self.leases[name],
            )
            for _id in ids:
                time_since_refresh = now - self.leases[name][_id]
                if time_since_refresh > self.lease_timeout:
                    logger.debug(
                        "Lease %s for %s timed out after %ss.",
                        _id,
                        name,
                        time_since_refresh,
                    )
                    self._release_value(name=name, lease_id=_id)

    def close(self, comm=None, name=None):
        """Hard close the semaphore without warning clients which still hold a lease."""
        with log_errors():
            if not self._semaphore_exists(name):
                return

            del self.max_leases[name]
            if name in self.events:
                del self.events[name]
            if name in self.leases:
                if self.leases[name]:
                    warnings.warn(
                        f"Closing semaphore {name} but there remain unreleased leases {sorted(self.leases[name])}",
                        RuntimeWarning,
                    )
                del self.leases[name]
            if name in self.metrics["pending"]:
                if self.metrics["pending"][name]:
                    warnings.warn(
                        f"Closing semaphore {name} but there remain pending leases",
                        RuntimeWarning,
                    )
            # Clean-up state of semaphore metrics
            for _, metric_dict in self.metrics.items():
                if name in metric_dict:
                    del metric_dict[name]


class Semaphore:
    """Semaphore

    This `semaphore <https://en.wikipedia.org/wiki/Semaphore_(programming)>`_
    will track leases on the scheduler which can be acquired and
    released by an instance of this class. If the maximum amount of leases are
    already acquired, it is not possible to acquire more and the caller waits
    until another lease has been released.

    The lifetime or leases are controlled using a timeout. This timeout is
    refreshed in regular intervals by the ``Client`` of this instance and
    provides protection from deadlocks or resource starvation in case of worker
    failure.
    The timeout can be controlled using the configuration option
    ``distributed.scheduler.locks.lease-timeout`` and the interval in which the
    scheduler verifies the timeout is set using the option
    ``distributed.scheduler.locks.lease-validation-interval``.

    A noticeable difference to the Semaphore of the python standard library is
    that this implementation does not allow to release more often than it was
    acquired. If this happens, a warning is emitted but the internal state is
    not modified.

    .. warning::

        This implementation is still in an experimental state and subtle
        changes in behavior may occur without any change in the major version
        of this library.

    .. warning::

        This implementation is susceptible to lease overbooking in case of
        lease timeouts. It is advised to monitor log information and adjust
        above configuration options to suitable values for the user application.

    Parameters
    ----------
    max_leases: int (optional)
        The maximum amount of leases that may be granted at the same time. This
        effectively sets an upper limit to the amount of parallel access to a specific resource.
        Defaults to 1.
    name: string (optional)
        Name of the semaphore to acquire.  Choosing the same name allows two
        disconnected processes to coordinate.  If not given, a random
        name will be generated.
    register: bool
        If True, register the semaphore with the scheduler. This needs to be
        done before any leases can be acquired. If not done during
        initialization, this can also be done by calling the register method of
        this class.
        When registering, this needs to be awaited.
    scheduler_rpc: ConnectionPool
        The ConnectionPool to connect to the scheduler. If None is provided, it
        uses the worker or client pool. This paramter is mostly used for
        testing.
    loop: IOLoop
        The event loop this instance is using. If None is provided, reuse the
        loop of the active worker or client.

    Examples
    --------
    >>> from distributed import Semaphore
    ... sem = Semaphore(max_leases=2, name='my_database')
    ...
    ... def access_resource(s, sem):
    ...     # This automatically acquires a lease from the semaphore (if available) which will be
    ...     # released when leaving the context manager.
    ...     with sem:
    ...         pass
    ...
    ... futures = client.map(access_resource, range(10), sem=sem)
    ... client.gather(futures)
    ... # Once done, close the semaphore to clean up the state on scheduler side.
    ... sem.close()

    Notes
    -----
    If a client attempts to release the semaphore but doesn't have a lease acquired, this will raise an exception.


    When a semaphore is closed, if, for that closed semaphore, a client attempts to:

    - Acquire a lease: an exception will be raised.
    - Release: a warning will be logged.
    - Close: nothing will happen.


    dask executes functions by default assuming they are pure, when using semaphore acquire/releases inside
    such a function, it must be noted that there *are* in fact side-effects, thus, the function can no longer be
    considered pure. If this is not taken into account, this may lead to unexpected behavior.

    """

    def __init__(
        self,
        max_leases=1,
        name=None,
        register=True,
        scheduler_rpc=None,
        loop=None,
    ):

        try:
            worker = get_worker()
            self.scheduler = scheduler_rpc or worker.scheduler
            self.loop = loop or worker.loop

        except ValueError:
            client = get_client()
            self.scheduler = scheduler_rpc or client.scheduler
            self.loop = loop or client.io_loop

        self.name = name or "semaphore-" + uuid.uuid4().hex
        self.max_leases = max_leases
        self.id = uuid.uuid4().hex
        self._leases = deque()

        self.refresh_leases = True

        self._registered = None
        if register:
            self._registered = self.register()

        # this should give ample time to refresh without introducing another
        # config parameter since this *must* be smaller than the timeout anyhow
        refresh_leases_interval = (
            parse_timedelta(
                dask.config.get("distributed.scheduler.locks.lease-timeout"),
                default="s",
            )
            / 5
        )
        pc = PeriodicCallback(
            self._refresh_leases, callback_time=refresh_leases_interval * 1000
        )
        self.refresh_callback = pc

        # Need to start the callback using IOLoop.add_callback to ensure that the
        # PC uses the correct event loop.
        self.loop.add_callback(pc.start)

    @property
    def asynchronous(self):
        return self.loop is IOLoop.current()

    async def _register(self):
        await retry_operation(
            self.scheduler.semaphore_register,
            name=self.name,
            max_leases=self.max_leases,
            operation=f"semaphore register id={self.id} name={self.name}",
        )

    def register(self, **kwargs):
        return self.sync(self._register)

    def __await__(self):
        async def create_semaphore():
            if self._registered:
                await self._registered
            return self

        return create_semaphore().__await__()

    def sync(self, func, *args, asynchronous=None, callback_timeout=None, **kwargs):
        callback_timeout = parse_timedelta(callback_timeout)
        if (
            asynchronous
            or self.asynchronous
            or getattr(thread_state, "asynchronous", False)
        ):
            future = func(*args, **kwargs)
            if callback_timeout is not None:
                future = asyncio.wait_for(future, callback_timeout)
            return future
        else:
            return sync(
                self.loop, func, *args, callback_timeout=callback_timeout, **kwargs
            )

    async def _refresh_leases(self):
        if self.refresh_leases and self._leases:
            logger.debug(
                "%s refreshing leases for %s with IDs %s",
                self.id,
                self.name,
                self._leases,
            )
            await retry_operation(
                self.scheduler.semaphore_refresh_leases,
                lease_ids=list(self._leases),
                name=self.name,
                operation="semaphore refresh leases: id=%s, lease_ids=%s, name=%s"
                % (self.id, list(self._leases), self.name),
            )

    async def _acquire(self, timeout=None):
        lease_id = uuid.uuid4().hex
        logger.debug(
            "%s requests lease for %s with ID %s", self.id, self.name, lease_id
        )

        # Using a unique lease id generated here allows us to retry since the
        # server handle is idempotent
        result = await retry_operation(
            self.scheduler.semaphore_acquire,
            name=self.name,
            timeout=timeout,
            lease_id=lease_id,
            operation="semaphore acquire: id=%s, lease_id=%s, name=%s"
            % (self.id, lease_id, self.name),
        )
        if result:
            self._leases.append(lease_id)
        return result

    def acquire(self, timeout=None):
        """
        Acquire a semaphore.

        If the internal counter is greater than zero, decrement it by one and return True immediately.
        If it is zero, wait until a release() is called and return True.

        Parameters
        ----------
        timeout : number or string or timedelta, optional
            Seconds to wait on acquiring the semaphore.  This does not
            include local coroutine time, network transfer time, etc..
            Instead of number of seconds, it is also possible to specify
            a timedelta in string format, e.g. "200ms".
        """
        timeout = parse_timedelta(timeout)
        return self.sync(self._acquire, timeout=timeout)

    async def _release(self, lease_id):
        try:
            await retry_operation(
                self.scheduler.semaphore_release,
                name=self.name,
                lease_id=lease_id,
                operation="semaphore release: id=%s, lease_id=%s, name=%s"
                % (self.id, lease_id, self.name),
            )
            return True
        except Exception:  # Release fails for whatever reason
            logger.error(
                "Release failed for id=%s, lease_id=%s, name=%s. Cluster network might be unstable?"
                % (self.id, lease_id, self.name),
                exc_info=True,
            )
            return False

    def release(self):
        """
        Release the semaphore.

        Returns
        -------
        bool
            This value indicates whether a lease was released immediately or not. Note that a user should  *not* retry
            this operation. Under certain circumstances (e.g. scheduler overload) the lease may not be released
            immediately, but it will always be automatically released after a specific interval configured using
            "distributed.scheduler.locks.lease-validation-interval" and "distributed.scheduler.locks.lease-timeout".
        """
        if not self._leases:
            raise RuntimeError("Released too often")

        # popleft to release the oldest lease first
        lease_id = self._leases.popleft()
        logger.debug("%s releases %s for %s", self.id, lease_id, self.name)
        return self.sync(self._release, lease_id=lease_id)

    def get_value(self):
        """
        Return the number of currently registered leases.
        """
        return self.sync(self.scheduler.semaphore_value, name=self.name)

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, *args, **kwargs):
        self.release()

    async def __aenter__(self):
        await self.acquire()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.release()

    def __getstate__(self):
        # Do not serialize the address since workers may have different
        # addresses for the scheduler (e.g. if a proxy is between them)
        return (self.name, self.max_leases)

    def __setstate__(self, state):
        name, max_leases = state
        self.__init__(
            name=name,
            max_leases=max_leases,
            register=False,
        )

    def close(self):
        return self.sync(self.scheduler.semaphore_close, name=self.name)

    def __del__(self):
        self.refresh_callback.stop()



import asyncio
import re
from operator import add, sub
from time import sleep

import pytest

pytest.importorskip("bokeh")
from tlz import first
from tornado.httpclient import AsyncHTTPClient

from distributed.client import wait
from distributed.dashboard.components.worker import (
    CommunicatingStream,
    CommunicatingTimeSeries,
    Counters,
    CrossFilter,
    ExecutingTimeSeries,
    StateTable,
    SystemMonitor,
)
from distributed.metrics import time
from distributed.utils_test import dec, gen_cluster, inc


@gen_cluster(
    client=True, worker_kwargs={"dashboard": True}, scheduler_kwargs={"dashboard": True}
)
async def test_routes(c, s, a, b):
    port = a.http_server.port

    future = c.submit(sleep, 1)
    await asyncio.sleep(0.1)

    http_client = AsyncHTTPClient()
    for suffix in ["status", "counters", "system", "profile", "profile-server"]:
        response = await http_client.fetch("http://localhost:%d/%s" % (port, suffix))
        body = response.body.decode()
        assert "bokeh" in body.lower()
        assert not re.search("href=./", body)  # no absolute links

    response = await http_client.fetch(
        "http://localhost:%d/info/main/workers.html" % s.http_server.port
    )

    assert str(port) in response.body.decode()


@gen_cluster(client=True, worker_kwargs={"dashboard": True})
async def test_simple(c, s, a, b):
    assert s.workers[a.address].services == {"dashboard": a.http_server.port}
    assert s.workers[b.address].services == {"dashboard": b.http_server.port}

    future = c.submit(sleep, 1)
    await asyncio.sleep(0.1)

    http_client = AsyncHTTPClient()
    for suffix in ["crossfilter", "system"]:
        response = await http_client.fetch(
            "http://localhost:%d/%s" % (a.http_server.port, suffix)
        )
        assert "bokeh" in response.body.decode().lower()


@gen_cluster(client=True, worker_kwargs={"dashboard": True})
async def test_services_kwargs(c, s, a, b):
    assert s.workers[a.address].services == {"dashboard": a.http_server.port}


@gen_cluster(client=True)
async def test_basic(c, s, a, b):
    for component in [
        StateTable,
        ExecutingTimeSeries,
        CommunicatingTimeSeries,
        CrossFilter,
        SystemMonitor,
    ]:

        aa = component(a)
        bb = component(b)

        xs = c.map(inc, range(10), workers=a.address)
        ys = c.map(dec, range(10), workers=b.address)

        def slowall(*args):
            sleep(1)

        x = c.submit(slowall, xs, ys, 1, workers=a.address)
        y = c.submit(slowall, xs, ys, 2, workers=b.address)
        await asyncio.sleep(0.1)

        aa.update()
        bb.update()

        assert len(first(aa.source.data.values())) and len(
            first(bb.source.data.values())
        )


@gen_cluster(client=True)
async def test_counters(c, s, a, b):
    pytest.importorskip("crick")
    while "tick-duration" not in a.digests:
        await asyncio.sleep(0.01)
    aa = Counters(a)

    aa.update()
    await asyncio.sleep(0.1)
    aa.update()

    start = time()
    while not len(aa.digest_sources["tick-duration"][0].data["x"]):
        await asyncio.sleep(1)
        assert time() < start + 5

    a.digests["foo"].add(1)
    a.digests["foo"].add(2)
    aa.add_digest_figure("foo")

    a.counters["bar"].add(1)
    a.counters["bar"].add(2)
    a.counters["bar"].add(2)
    aa.add_counter_figure("bar")

    for x in [aa.counter_sources.values(), aa.digest_sources.values()]:
        for y in x:
            for z in y.values():
                assert len(set(map(len, z.data.values()))) == 1


@gen_cluster(client=True)
async def test_CommunicatingStream(c, s, a, b):
    aa = CommunicatingStream(a)
    bb = CommunicatingStream(b)

    xs = c.map(inc, range(10), workers=a.address)
    ys = c.map(dec, range(10), workers=b.address)
    adds = c.map(add, xs, ys, workers=a.address)
    subs = c.map(sub, xs, ys, workers=b.address)

    await wait([adds, subs])

    aa.update()
    bb.update()

    assert len(first(aa.outgoing.data.values())) and len(
        first(bb.outgoing.data.values())
    )
    assert len(first(aa.incoming.data.values())) and len(
        first(bb.incoming.data.values())
    )


@gen_cluster(
    client=True, clean_kwargs={"threads": False}, worker_kwargs={"dashboard": True}
)
async def test_prometheus(c, s, a, b):
    pytest.importorskip("prometheus_client")

    http_client = AsyncHTTPClient()
    for suffix in ["metrics"]:
        response = await http_client.fetch(
            "http://localhost:%d/%s" % (a.http_server.port, suffix)
        )
        assert response.code == 200



import asyncio
from time import sleep

from distributed import MultiLock, get_client
from distributed.metrics import time
from distributed.multi_lock import MultiLockExtension
from distributed.utils_test import (  # noqa F401
    client,
    cluster_fixture,
    gen_cluster,
    loop,
)


@gen_cluster(client=True, nthreads=[("127.0.0.1", 8)] * 2)
async def test_single_lock(c, s, a, b):
    await c.set_metadata("locked", False)

    def f(_):
        client = get_client()
        with MultiLock(names=["x"]):
            assert client.get_metadata("locked") is False
            client.set_metadata("locked", True)
            sleep(0.05)
            assert client.get_metadata("locked") is True
            client.set_metadata("locked", False)

    futures = c.map(f, range(20))
    await c.gather(futures)
    ext: MultiLockExtension = s.extensions["multi_locks"]
    assert not ext.events
    assert not ext.requests
    assert not ext.requests_left
    assert all(len(l) == 0 for l in ext.locks.values())


@gen_cluster(client=True)
async def test_timeout(c, s, a, b):
    ext: MultiLockExtension = s.extensions["multi_locks"]
    lock1 = MultiLock(names=["x"])
    result = await lock1.acquire()
    assert result is True
    assert ext.requests_left[lock1.id] == 0
    assert ext.locks["x"] == [lock1.id]
    assert not ext.events

    lock2 = MultiLock(names=["x"])
    assert lock1.id != lock2.id

    start = time()
    result = await lock2.acquire(timeout=0.1)
    stop = time()
    assert stop - start < 0.3
    assert result is False
    assert ext.locks["x"] == [lock1.id]
    assert not ext.events

    await lock1.release()


@gen_cluster(client=True)
async def test_timeout_wake_waiter(c, s, a, b):
    ext: MultiLockExtension = s.extensions["multi_locks"]
    l1 = MultiLock(names=["x"])
    l2 = MultiLock(names=["x", "y"])
    l3 = MultiLock(names=["y"])
    await l1.acquire()

    l2_acquire = asyncio.ensure_future(l2.acquire(timeout=1))
    try:
        await asyncio.wait_for(asyncio.shield(l2_acquire), 0.1)
    except asyncio.TimeoutError:
        pass
    else:
        assert False

    l3_acquire = asyncio.ensure_future(l3.acquire())
    try:
        await asyncio.wait_for(asyncio.shield(l3_acquire), 0.1)
    except asyncio.TimeoutError:
        pass
    else:
        assert False

    assert await l2_acquire is False
    assert await l3_acquire
    l1.release()
    l3.release()


@gen_cluster(client=True)
async def test_multiple_locks(c, s, a, b):
    ext: MultiLockExtension = s.extensions["multi_locks"]
    l1 = MultiLock(names=["l1"])
    l2 = MultiLock(names=["l2"])
    l3 = MultiLock(names=["l1", "l2"])

    # Both `l1` and `l2` are free to acquire
    assert await l1.acquire()
    assert await l2.acquire()
    assert list(ext.locks.keys()) == ["l1", "l2"]
    assert list(ext.locks.values()) == [[l1.id], [l2.id]]
    assert list(ext.requests.keys()) == [l1.id, l2.id]
    assert list(ext.requests_left.values()) == [0, 0]
    assert not ext.events

    # Since `l3` requires both `l1` and `l2`, it isn't available immediately
    l3_acquire = asyncio.ensure_future(l3.acquire())
    try:
        await asyncio.wait_for(asyncio.shield(l3_acquire), 0.1)
    except asyncio.TimeoutError:
        assert list(ext.locks.keys()) == ["l1", "l2"]
        assert list(ext.locks.values()) == [[l1.id, l3.id], [l2.id, l3.id]]
        assert ext.requests[l3.id] == {"l1", "l2"}
        assert ext.requests_left[l3.id] == 2
        assert l3.id in ext.events
    else:
        assert False  # We except a TimeoutError since `l3` isn't availabe

    # Releasing `l1` isn't enough since `l3` also requires `l2`
    await l1.release()
    try:
        await asyncio.wait_for(asyncio.shield(l3_acquire), 0.1)
    except asyncio.TimeoutError:
        # `l3` now only wait on `l2`
        assert list(ext.locks.keys()) == ["l1", "l2"]
        assert list(ext.locks.values()) == [[l3.id], [l2.id, l3.id]]
        assert ext.requests[l3.id] == {"l1", "l2"}
        assert ext.requests_left[l3.id] == 1
        assert l3.id in ext.events
    else:
        assert False

    # Releasing `l2` should make `l3` available
    await l2.release()
    assert list(ext.locks.keys()) == ["l1", "l2"]
    assert list(ext.locks.values()) == [[l3.id], [l3.id]]
    assert ext.requests[l3.id] == {"l1", "l2"}
    assert ext.requests_left[l3.id] == 0

    await l3.release()
    assert not ext.events
    assert not ext.requests
    assert not ext.requests_left
    assert all(len(l) == 0 for l in ext.locks.values())


@gen_cluster(client=True)
async def test_num_locks(c, s, a, b):
    ext: MultiLockExtension = s.extensions["multi_locks"]
    l1 = MultiLock(names=["l1", "l2", "l3"])
    l2 = MultiLock(names=["l1", "l2", "l3"])
    l3 = MultiLock(names=["l1", "l2", "l3", "l4"])

    # Even though `l1` and `l2` uses the same lock names they
    # only requires a subset of the locks
    assert await l1.acquire(num_locks=1)
    assert await l2.acquire(num_locks=2)
    assert list(ext.locks.keys()) == ["l1", "l2", "l3"]
    assert list(ext.locks.values()) == [[l1.id], [l2.id], [l2.id]]
    assert list(ext.requests.keys()) == [l1.id, l2.id]
    assert list(ext.requests_left.values()) == [0, 0]
    assert not ext.events

    # Since `l3` requires three out of four locks it has to wait
    l3_acquire = asyncio.ensure_future(l3.acquire(num_locks=3))
    try:
        await asyncio.wait_for(asyncio.shield(l3_acquire), 0.1)
    except asyncio.TimeoutError:
        assert list(ext.locks.keys()) == ["l1", "l2", "l3", "l4"]
        assert list(ext.locks.values()) == [
            [l1.id, l3.id],
            [l2.id, l3.id],
            [l2.id, l3.id],
            [l3.id],
        ]
        assert list(ext.requests_left.values()) == [0, 0, 2]
        assert l3.id in ext.events
    else:
        assert False  # We except a TimeoutError since `l3` isn't availabe

    # Releasing `l1` isn't enough since `l3` also requires three locks
    await l1.release()
    try:
        await asyncio.wait_for(asyncio.shield(l3_acquire), 0.1)
    except asyncio.TimeoutError:
        assert list(ext.locks.keys()) == ["l1", "l2", "l3", "l4"]
        assert list(ext.locks.values()) == [
            [l3.id],
            [l2.id, l3.id],
            [l2.id, l3.id],
            [l3.id],
        ]
        assert list(ext.requests.keys()) == [l2.id, l3.id]
        assert list(ext.requests_left.values()) == [0, 1]
        assert l3.id in ext.events
    else:
        assert False

    # Releasing `l2` is enough to release `l3`
    await l2.release()
    await asyncio.sleep(0.1)  # Give `l3` a change to wake up and acquire its locks
    assert list(ext.locks.keys()) == ["l1", "l2", "l3", "l4"]
    assert list(ext.locks.values()) == [[l3.id], [l3.id], [l3.id], [l3.id]]
    assert list(ext.requests.keys()) == [l3.id]
    assert list(ext.requests_left.values()) == [0]
    assert l3.id not in ext.events

    await l3.release()
    assert not ext.events
    assert not ext.requests
    assert not ext.requests_left
    assert all(len(l) == 0 for l in ext.locks.values())



from time import sleep

import pytest

import dask
from dask.utils import format_bytes

from distributed import Client
from distributed.comm.ucx import _scrub_ucx_config
from distributed.utils import get_ip
from distributed.utils_test import cleanup, gen_test, inc, loop, popen  # noqa: 401

try:
    HOST = get_ip()
except Exception:
    HOST = "127.0.0.1"

ucp = pytest.importorskip("ucp")
rmm = pytest.importorskip("rmm")


@pytest.mark.asyncio
async def test_ucx_config(cleanup):

    ucx = {
        "nvlink": True,
        "infiniband": True,
        "rdmacm": False,
        "net-devices": "",
        "tcp": True,
        "cuda_copy": True,
    }

    with dask.config.set(ucx=ucx):
        ucx_config = _scrub_ucx_config()
        assert ucx_config.get("TLS") == "rc,tcp,sockcm,cuda_copy,cuda_ipc"
        assert ucx_config.get("SOCKADDR_TLS_PRIORITY") == "sockcm"
        assert ucx_config.get("NET_DEVICES") is None

    ucx = {
        "nvlink": False,
        "infiniband": True,
        "rdmacm": False,
        "net-devices": "mlx5_0:1",
        "tcp": True,
        "cuda_copy": False,
    }

    with dask.config.set(ucx=ucx):
        ucx_config = _scrub_ucx_config()
        assert ucx_config.get("TLS") == "rc,tcp,sockcm"
        assert ucx_config.get("SOCKADDR_TLS_PRIORITY") == "sockcm"
        assert ucx_config.get("NET_DEVICES") == "mlx5_0:1"

    ucx = {
        "nvlink": False,
        "infiniband": True,
        "rdmacm": True,
        "net-devices": "all",
        "MEMTYPE_CACHE": "y",
        "tcp": True,
        "cuda_copy": True,
    }

    with dask.config.set(ucx=ucx):
        ucx_config = _scrub_ucx_config()
        assert ucx_config.get("TLS") == "rc,tcp,rdmacm,cuda_copy"
        assert ucx_config.get("SOCKADDR_TLS_PRIORITY") == "rdmacm"
        assert ucx_config.get("MEMTYPE_CACHE") == "y"


def test_ucx_config_w_env_var(cleanup, loop, monkeypatch):
    size = "1000.00 MB"
    monkeypatch.setenv("DASK_RMM__POOL_SIZE", size)

    dask.config.refresh()

    port = "13339"
    sched_addr = "ucx://%s:%s" % (HOST, port)

    with popen(
        ["dask-scheduler", "--no-dashboard", "--protocol", "ucx", "--port", port]
    ) as sched:
        with popen(
            [
                "dask-worker",
                sched_addr,
                "--no-dashboard",
                "--protocol",
                "ucx",
                "--no-nanny",
            ]
        ) as w:
            with Client(sched_addr, loop=loop, timeout=10) as c:
                while not c.scheduler_info()["workers"]:
                    sleep(0.1)

                # configured with 1G pool
                rmm_usage = c.run_on_scheduler(rmm.get_info)
                assert size == format_bytes(rmm_usage.free)

                # configured with 1G pool
                worker_addr = list(c.scheduler_info()["workers"])[0]
                worker_rmm_usage = c.run(rmm.get_info)
                rmm_usage = worker_rmm_usage[worker_addr]
                assert size == format_bytes(rmm_usage.free)



from . import config  # isort:skip; load distributed configuration first
import dask
from dask.config import config
from dask.utils import import_required

from ._version import get_versions
from .actor import Actor, ActorFuture
from .client import (
    Client,
    CompatibleExecutor,
    Executor,
    Future,
    as_completed,
    default_client,
    fire_and_forget,
    futures_of,
    get_task_metadata,
    get_task_stream,
    performance_report,
    wait,
)
from .core import Status, connect, rpc
from .deploy import Adaptive, LocalCluster, SpecCluster, SSHCluster
from .diagnostics.plugin import PipInstall, SchedulerPlugin, WorkerPlugin
from .diagnostics.progressbar import progress
from .event import Event
from .lock import Lock
from .multi_lock import MultiLock
from .nanny import Nanny
from .pubsub import Pub, Sub
from .queues import Queue
from .scheduler import Scheduler
from .security import Security
from .semaphore import Semaphore
from .threadpoolexecutor import rejoin
from .utils import CancelledError, TimeoutError, sync
from .variable import Variable
from .worker import Reschedule, Worker, get_client, get_worker, secede
from .worker_client import local_client, worker_client

versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

if dask.config.get("distributed.admin.event-loop") in ("asyncio", "tornado"):
    pass
elif dask.config.get("distributed.admin.event-loop") == "uvloop":
    import_required(
        "uvloop",
        "The distributed.admin.event-loop configuration value "
        "is set to 'uvloop' but the uvloop module is not installed"
        "\n\n"
        "Please either change the config value or install one of the following\n"
        "    conda install uvloop\n"
        "    pip install uvloop",
    )
    import uvloop

    uvloop.install()
else:
    raise ValueError(
        "Expected distributed.admin.event-loop to be in ('asyncio', 'tornado', 'uvloop'), got %s"
        % dask.config.get("distributed.admin.event-loop")
    )



import pytest

from distributed.protocol import deserialize, serialize

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")


@pytest.mark.parametrize("collection", [tuple, dict, list])
@pytest.mark.parametrize(
    "y,y_serializer",
    [
        (np.arange(50), "dask"),
        (pd.DataFrame({"C": ["a", "b", None], "D": [2.5, 3.5, 4.5]}), "pickle"),
        (None, "pickle"),
    ],
)
def test_serialize_collection(collection, y, y_serializer):
    x = np.arange(100)
    if issubclass(collection, dict):
        header, frames = serialize({"x": x, "y": y}, serializers=("dask", "pickle"))
    else:
        header, frames = serialize(collection((x, y)), serializers=("dask", "pickle"))
    frames = tuple(frames)  # verify that no mutation occurs
    t = deserialize(header, frames, deserializers=("dask", "pickle", "error"))
    assert isinstance(t, collection)

    assert header["is-collection"] is True
    sub_headers = header["sub-headers"]

    if collection is not dict:
        assert sub_headers[0]["serializer"] == "dask"
        assert sub_headers[1]["serializer"] == y_serializer

    if collection is dict:
        assert (t["x"] == x).all()
        assert str(t["y"]) == str(y)
    else:
        assert (t[0] == x).all()
        assert str(t[1]) == str(y)


def test_large_collections_serialize_simply():
    header, frames = serialize(tuple(range(1000)))
    assert len(frames) == 1


def test_nested_types():
    x = np.ones(5)
    header, frames = serialize([[[x]]])
    assert "dask" in str(header)
    assert len(frames) == 1
    assert x.data == np.frombuffer(frames[0]).data
